"""Configuration management utilities"""
import os, time
import json
import types
import threading
from ..common import getLogger


__copyright__ = "Copyright 2023 binchoo"
__license__ = "GPLv3"
__author__ = "Jaebin Joo"
__email__ = "jaebin.joo@megazone.com"


_logger_ = getLogger(__name__)


def import_config(path: str):
    '''
    Reads json-formatted file and creates a configuration object from it
    :return: config
    '''
    try:
        with open(path, 'r') as fp:
            config = json.load(fp, object_hook=lambda x: types.SimpleNamespace(**x))
            return config
    except json.JSONDecodeError:
        _logger_.error(f"{path} has invalid format")
    except Exception:
        _logger_.error(f"{path} is not readable")
    return None


class ConfigLoader:

    def __init__(self, path: str, interval: float):
        '''Creates ConfigLoader.
        It periodically detects modification of the configuration file, and updates values.
        :param path: configuration file path
        :param interval: hot replacement interval
        '''
        self.path = path
        self.config = import_config(path)
        self.modified_time = os.path.getmtime(self.path)
        self.dirty = True

        self.previous_config = self.config

        self.interval = min(interval, 60)
        self.runnable = True

    def keep_replacing(self):
        while self.runnable:
            if self.is_modified():
                self.replace_config()
            time.sleep(self.interval)

    def is_modified(self):
        modified = os.path.getmtime(self.path) != self.modified_time
        if modified:
            _logger_.debug("Change detected")
        else:
            _logger_.debug("Change not detected")
        return modified

    def replace_config(self):
        self.previous_config = self.config
        self.config = import_config(self.path)
        self.modified_time = os.path.getmtime(self.path)

        _logger_.info(f"Configuration has been changed\n"
                      f"\tFrom: {self.previous_config}\n"
                      f"\tTo: {self.config}")

    def run(self):
        hot_replacer = threading.Thread(target=self.keep_replacing)
        hot_replacer.start()

    def close(self):
        self.runnable = False
        _logger_.debug("ConfigLoader will stop")

    def read(self, keys=None, config=None):
        obj = self.config if config is None else config
        if keys is None or len(keys) == 0:
            return obj
        else:
            for i, key in enumerate(keys):
                if hasattr(obj, key):
                    if i + 1 == len(keys):
                        return getattr(obj, key)
                    else:
                        obj = getattr(obj, key)
                else:
                    return None
            return obj

    @DeprecationWarning
    def is_key_modified(self, keys=None):
        current_value = self.read(keys, self.config)
        previous_value = self.read(keys, self.previous_config)

        if hasattr(current_value, '__dict__') \
                and hasattr(previous_value, '__dict__'):

            return not (current_value.__dict__ == previous_value.__dict__)

        elif not hasattr(current_value, '__dict__') \
                and not hasattr(previous_value, '__dict__'):

            return current_value == previous_value

        return False


def automatic_loader(path: str='config.json', poll_interval: float = 5) -> ConfigLoader:
    '''
    :param poll_interval: poll interval seconds
    :return: ConfigLoader
    '''
    config_loader = ConfigLoader(path, poll_interval)
    config_loader.run()
    return config_loader



"""Provides configured logger"""
import logging


formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

log_handler = logging.StreamHandler()
log_handler.setFormatter(formatter)

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.INFO)
logger.addHandler(log_handler)

children = []

def getLogger(name: str, level: int | str=logging.INFO):
    child_logger = logger.getChild(suffix=name)
    child_logger.setLevel(level)
    children.append(child_logger)
    return child_logger

def setLogLevel(level):
    for child_logger in children:
        child_logger.setLevel(level)
    logger.setLevel(level)



import logging
import grasshopper
from awsiot.greengrasscoreipc.clientv2 import GreengrassCoreIPCClientV2
from grasshopper.config import automatic_loader
from grasshopper.modbus import ModbusTCPConnection, GGv2ModbusConnector


logger = grasshopper.common.getLogger(__name__)
grasshopper.common.setLogLevel(level=logging.INFO)


if __name__ == '__main__':
    connector = GGv2ModbusConnector(
        ipc=None,
        conn=ModbusTCPConnection(),
        cfg=automatic_loader())

    try:
        connector.start() # synchronous
    except Exception as e:
        logger.exception(e)
    finally:
        connector.close()



"""Modbus connection management"""
import time
from pymodbus.client import ModbusTcpClient
from ..common import getLogger


__copyright__ = "Copyright 2023 binchoo"
__license__ = "GPLv3"
__author__ = "Jaebin Joo"
__email__ = "jaebin.joo@megazone.com"


_logger_ = getLogger(__name__)

def safe(getter, default):
    if getter is not None:
        v = getter()
        return v if v is not None else default
    return default


class SingleModbusConnection:
    '''Singleton modbus connection manager'''

    def get(self, *args, **kwargs):
        raise NotImplementedError


class ModbusTCPConnection(SingleModbusConnection):

    DEFAULT_HOST = '127.0.0.1'
    DEFAULT_PORT = 502
    DEFAULT_WAIT_SEC = 10
    DEFAULT_MAX_RETRY = 2147483647

    def __init__(self):
        self.client = None # singleton connection

    def get(self, get_host, get_port, get_max_retry=None, get_wait_sec=None, **kwargs):
        max_retry = max(0, safe(get_max_retry, self.DEFAULT_MAX_RETRY))
        while max_retry >= 0:
            host = safe(get_host, self.DEFAULT_HOST)
            port = safe(get_port, self.DEFAULT_PORT)
            wait_sec = safe(get_wait_sec, self.DEFAULT_WAIT_SEC)

            _logger_.info(f"Connecting to {host}:{port} (max_retry={max_retry}, wait_sec={wait_sec})")

            if self._connect(host, port):
                return self.client

            max_retry -= 1
            time.sleep(wait_sec)

        raise RuntimeError("Connection ultimately failed")

    def _connect(self, host, port):
        if self.client is not None:
            self.client.close()

        self.client = ModbusTcpClient(host=host, port=port)
        self.client.connect()

        if self.client.connected:
            _logger_.info(f"Connected to {host}:{port}")
            return True
        else:
            _logger_.warning(f"Connection failed")
            return False

class ModbusRTUConnection(SingleModbusConnection):
    pass



"""AWS IoT Greengrass custom modbus connector"""
import time
import json
from awsiot.greengrasscoreipc.clientv2 import GreengrassCoreIPCClientV2
from awsiot.greengrasscoreipc.model import QOS, PayloadFormat, PublishMessage, JsonMessage
from ..config.configloader import ConfigLoader
from ..common import getLogger
from .connection import SingleModbusConnection

__copyright__ = "Copyright 2023 binchoo"
__license__ = "GPLv3"
__author__ = "Jaebin Joo"
__email__ = "jaebin.joo@megazone.com"


_logger_ = getLogger(__name__)


class GGv2ModbusConnector:

    AWS_IOT_CORE = 'IotCore'
    LOCAL_PUBSUB = 'PubSub'

    def __init__(self, ipc: GreengrassCoreIPCClientV2,
                 conn: SingleModbusConnection, cfg: ConfigLoader, callback=None):
        '''Creates a GGv2ModbusConnector.This component reads modbus holding registers
        and send values to greengrass local PubSub.
        The runtime **must provide** AWS IoT Greengrass V2 Nucleus, local PubSub, and MQTT Bridge.
        Current component **must be privileged** in use of those GGv2 components.

        :param ipc: AWS IoT GGv2 IPC client
        :param conn: SingleModbusConnection that creates modbus connection to host:port
        :param cfg: ConfigLoader that provides configuration value
        '''
        self.ggv2_ipc = ipc
        self.connection = conn
        self.config_loader = cfg
        self.modbus_client = None
        self.callback = callback

    def start(self):
        '''Periodically retrieves holding registers from server.
        Values will soon be sent to Greengrass local PubSub via IPC.
        When appropriate configuration not provided, it talks to 127.0.0.1:502 by default.

        :raises RuntimeError when connection finally not established
        '''
        while True:
            self.setup_modbus()
            registers = self._read_holding_registers()
            if registers is not None:
                self._publish_registers(registers)
            self._wait()

    def setup_modbus(self):
        '''Get connected to modbus server'''
        get_host = lambda: self.config_loader.read(['server', 'host'])
        get_port = lambda: self.config_loader.read(['server', 'port'])
        get_max_retry = lambda: self.config_loader.read(['server', 'max_retry'])
        get_wait_sec = lambda: self.config_loader.read(['server', 'wait_sec'])

        if not self.check_modbus() \
                or (get_host() != self.modbus_client.comm_params.host) \
                or (get_port() != self.modbus_client.comm_params.port):

            self.close_modbus()
            self.modbus_client = self.connection \
                .get(get_host, get_port, get_max_retry, get_wait_sec)

    def close_modbus(self):
        '''Cleans-up modbus connection'''
        if self.modbus_client is not None:
            self.modbus_client.close()
        self.modbus_client = None

    def check_modbus(self):
        '''Verify connection is non-null and actually connected to the server
        :return: True if connection is established else False
        '''
        return self.modbus_client is not None \
            and self.modbus_client.connected

    def _read_holding_registers(self):
        '''Read holding registers from server

        :raises RuntimeError when connection is invalid
        :raises Exception when impossible in reading registers
        '''
        registers = None

        try:
            slave = max(0, self.config_loader.read(['holding_register', 'slave']))
            start, end_exclusive = self.config_loader.read(['holding_register', 'range'])
            rr = self.modbus_client.read_holding_registers(address=start, count=end_exclusive - start, slave=slave)
            registers = rr.registers
        except:
            self.close_modbus()
            _logger_.warning("Failed to read holding registers")
        return registers

    def _parse(self, registers, mapping):
        data = {}
        for attr, index in zip(mapping.attributes, mapping.indices):
            try:
                data[attr] = registers[index]
            except Exception as e:
                _logger_.exception(e)
        return data

    def _publish_registers(self, registers):
        '''Publish to local PubSub after converting registers in json-formatted data\n
        :param registers: values retrieved from server
        '''
        try:
            iot_mappings = self.config_loader.read(['iot_mapping'])
            for place, mapping in iot_mappings.__dict__.items():
                target = getattr(mapping, 'target')
                topic = getattr(mapping, 'topic')
                model = self._parse(registers, mapping)
                self._publish_to(target, topic, model)
        except Exception as e:
            _logger_.warning(e)

    def _publish_to(self, target, topic, model):
        if self.ggv2_ipc is not None:
            _logger_.debug(f"{time.ctime()} {topic}: {model} via {target}")
            if self.AWS_IOT_CORE == target:
                message = json.dumps(model)
                self.ggv2_ipc.publish_to_iot_core_async(topic_name=topic, qos=QOS.AT_LEAST_ONCE,
                                                        payload=message, payload_format=PayloadFormat.UTF8)
            elif self.LOCAL_PUBSUB == target:
                message = PublishMessage(json_message=JsonMessage(message=model))
                self.ggv2_ipc.publish_to_topic_async(topic=topic, publish_message=message)
            else:
                _logger_.warning(f"Wrong target destination: {target}")
        else:
            _logger_.warning("IPC client is missing")
        self._run_callback(self.callback, model)

    def _run_callback(self, func, model):
        if func is not None:
            func(model)

    def _wait(self):
        wait_sec = max(0, self.config_loader.read(['client', 'wait_sec']))
        time.sleep(wait_sec)

    def close(self):
        self.close_modbus()
        if self.ggv2_ipc is not None:
            self.ggv2_ipc.close()
        if self.config_loader is not None:
            self.config_loader.close()
